

/*
 * Author: Tomi Jylhä-Ollila, Finland 2010-2018
 *
 * This file is part of Kunquat.
 *
 * CC0 1.0 Universal, http://creativecommons.org/publicdomain/zero/1.0/
 *
 * To the extent possible under law, Kunquat Affirmers have waived all
 * copyright and related or neighboring rights to Kunquat.
 */


#ifndef KQT_RANDOM_H
#define KQT_RANDOM_H


#include <decl.h>

#include <stdint.h>
#include <stdlib.h>


/**
 * The maximum length of a random context description string.
 */
#define CONTEXT_LEN_MAX 15


/**
 * This is a portable pseudo-random generator.
 */
struct Random
{
    char context[CONTEXT_LEN_MAX + 1];
    uint64_t seed;
    uint64_t state;
};


#define RANDOM_AUTO (&(Random){ .context = "", .seed = 0, .state = 0 })


/**
 * The maximum 32-bit number generated by the pseudo-random generator.
 */
#define KQT_RANDOM32_MAX UINT32_MAX


/**
 * The maximum 64-bit number generated by the pseudo-random generator.
 */
#define KQT_RANDOM64_MAX UINT64_MAX


/**
 * Initialise the Random generator.
 *
 * \param random    The Random generator -- must not be \c NULL.
 * \param context   The context description -- must not be \c NULL or longer
 *                  than \c CONTEXT_LEN_MAX bytes (excluding the null
 *                  terminator).
 *
 * \return   The parameter \a random.
 */
Random* Random_init(Random* random, const char* context);


/**
 * Set the random seed in the Random.
 *
 * \param random   The Random generator -- must not be \c NULL.
 * \param seed     The random seed.
 */
void Random_set_seed(Random* random, uint64_t seed);


/**
 * Restart the random sequence in the Random.
 *
 * \param random   The Random generator -- must not be \c NULL.
 */
void Random_reset(Random* random);


/**
 * Get a 32-bit integer from the Random generator.
 *
 * This function returns the 32 most significant bits of the 64-bit
 * state and thus have higher overall quality.
 *
 * \param random   The Random generator -- must not be \c NULL.
 *
 * \return   A pseudorandom integer in the range [0, KQT_RANDOM32_MAX].
 */
uint32_t Random_get_uint32(Random* random);


/**
 * Get a 64-bit integer from the Random generator.
 *
 * This function is generally not recommended unless a large number of
 * random bits is needed.
 *
 * \param random   The Random generator -- must not be \c NULL.
 *
 * \return   A pseudorandom integer in the range [0, KQT_RANDOM64_MAX].
 */
uint64_t Random_get_uint64(Random* random);


/**
 * Get a floating point number in the range [0, 1).
 *
 * \param random   The Random generator -- must not be \c NULL.
 *
 * \return   A pseudorandom floating-point number in the range [0, 1).
 */
double Random_get_float_lb(Random* random);


/**
 * Get an array index.
 *
 * \param random   The Random generator -- must not be \c NULL.
 * \param size     The array size -- must be > \c 0.
 *
 * \return   A pseudorandom integer in the range [0, \a size).
 */
int32_t Random_get_index(Random* random, int32_t size);


/**
 * Get a floating point number in the range [0, 1.0].
 *
 * This function is not suitable for converting to integers because it may
 * return exactly 1.0 as a result.
 *
 * \param random   The Random generator -- must not be \c NULL.
 *
 * \return   A pseudorandom floating-point number in the range [0, 1.0].
 */
double Random_get_float_scale(Random* random);


/**
 * Get a floating point number in the range [-1.0, 1.0].
 *
 * \param random   The Random generator -- must not be \c NULL.
 *
 * \return   A pseudorandom floating-point number in the range [-1.0, 1.0].
 */
double Random_get_float_signal(Random* random);


#endif // KQT_RANDOM_H


